#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library

# import interfaces
from pyams_workflow.interfaces import MANUAL, SYSTEM, AUTOMATIC, IWorkflow, InvalidTransitionError, IWorkflowState, \
    IWorkflowVersions, IWorkflowVersion, IWorkflowInfo, IWorkflowManagedContent, WorkflowVersionTransitionEvent, \
    WorkflowTransitionEvent, ConditionFailedError, NoTransitionAvailableError, AmbiguousTransitionError

# import packages
from pyams_utils.adapter import adapter_config
from pyams_utils.registry import get_utility
from pyams_utils.request import check_request
from pyams_utils.traversing import get_parent
from pyramid.httpexceptions import HTTPUnauthorized
from pyramid.threadlocal import get_current_registry
from zope.componentvocabulary.vocabulary import UtilityVocabulary
from zope.interface import implementer
from zope.lifecycleevent import ObjectModifiedEvent
from zope.schema.vocabulary import getVocabularyRegistry

from pyams_workflow import _


def NullCondition(wf, context):
    """Null condition"""
    return True


def NullAction(wf, context):
    """Null action"""
    pass


def granted_permission(permission, context):
    return True


class Transition(object):
    """Transition object

    A transition doesn't make anything by itself.
    Everything is handled by the workflow utility
    """

    def __init__(self, transition_id, title, source, destination,
                 condition=NullCondition,
                 action=NullAction,
                 trigger=MANUAL,
                 permission=None,
                 order=0,
                 **user_data):
        self.transition_id = transition_id
        self.title = title
        self.source = source
        self.destination = destination
        self.condition = condition
        self.action = action
        self.trigger = trigger
        self.permission = permission
        self.order = order
        self.user_data = user_data


@implementer(IWorkflow)
class Workflow(object):
    """Workflow utility"""

    def __init__(self, transitions, states, published_states=None):
        self.refresh(transitions)
        self.states = states
        self.published_states = published_states or set()

    def _register(self, transition):
        transitions = self._sources.setdefault(transition.source, {})
        transitions[transition.transition_id] = transition
        self._id_transitions[transition.transition_id] = transition

    def refresh(self, transitions):
        self._sources = {}
        self._id_transitions = {}
        for transition in transitions:
            self._register(transition)

    def get_state_label(self, state):
        try:
            return self.states.getTerm(state).title
        except LookupError:
            return _('-- unknown --')

    def get_transitions(self, source):
        try:
            return self._sources[source].values()
        except KeyError:
            return []

    def get_transition(self, source, transition_id):
        transition = self._id_transitions[transition_id]
        if transition.source != source:
            raise InvalidTransitionError(source)
        return transition

    def get_transition_by_id(self, transition_id):
        return self._id_transitions[transition_id]


WORKFLOW_STATE_KEY = 'pyams_workflow.state'


@adapter_config(context=IWorkflowVersion, provides=IWorkflowInfo)
class WorkflowInfo(object):
    """Workflow info adapter"""

    def __init__(self, context):
        self.context = context
        self.wf = get_utility(IWorkflow, name=self.name)

    @property
    def parent(self):
        return get_parent(self.context, IWorkflowManagedContent)

    @property
    def name(self):
        return self.parent.workflow_name

    def fire_transition(self, transition_id, comment=None, side_effect=None, check_security=True):
        versions = IWorkflowVersions(self.parent)
        state = IWorkflowState(self.context)
        # this raises InvalidTransitionError if id is invalid for current state
        transition = self.wf.get_transition(state.state, transition_id)
        # check whether we may execute this workflow transition
        if check_security and transition.permission:
            request = check_request()
            if not request.has_permission(transition.permission):
                raise HTTPUnauthorized()
        # now make sure transition can still work in this context
        if not transition.condition(self, self.context):
            raise ConditionFailedError()
        # perform action, return any result as new version
        result = transition.action(self, self.context)
        if result is not None:
            # clear result history
            IWorkflowState(result).history.clear()
            # stamp it with version
            versions.add_version(result, transition.destination)
            # execute any side effect:
            if side_effect is not None:
                side_effect(result)
            event = WorkflowVersionTransitionEvent(result, self.wf, self.context,
                                                   transition.source,
                                                   transition.destination,
                                                   transition, comment)
        else:
            versions.set_state(state.version_id, transition.destination)
            # execute any side effect
            if side_effect is not None:
                side_effect(self.context)
            event = WorkflowTransitionEvent(self.context, self.wf,
                                            transition.source,
                                            transition.destination,
                                            transition, comment)
        # change state of context or new object
        registry = get_current_registry()
        registry.notify(event)
        # send modified event for original or new object
        if result is None:
            registry.notify(ObjectModifiedEvent(self.context))
        else:
            registry.notify(ObjectModifiedEvent(result))
        return result

    def fire_transition_toward(self, state, comment=None, side_effect=None, check_security=True):
        transition_ids = self.get_fireable_transition_ids_toward(state, check_security)
        if not transition_ids:
            raise NoTransitionAvailableError(self.state(self.context).get_state(), state)
        if len(transition_ids) != 1:
            raise AmbiguousTransitionError(self.state(self.context).get_state(), state)
        return self.fire_transition(transition_ids[0], comment, side_effect, check_security)

    def fire_transition_for_versions(self, state, transition_id, comment=None):
        versions = IWorkflowVersions(self.parent)
        for version in versions.get_versions(state):
            IWorkflowInfo(version).fire_transition(transition_id, comment)

    def fire_automatic(self):
        for transition_id in self.get_automatic_transition_ids():
            try:
                self.fire_transition(transition_id)
            except ConditionFailedError:
                # if condition failed, that's fine, then we weren't
                # ready to fire yet
                pass
            else:
                # if we actually managed to fire a transition,
                # we're done with this one now.
                return

    def has_version(self, state):
        wf_versions = IWorkflowVersions(self.parent)
        return wf_versions.has_version(state)

    def get_manual_transition_ids(self, check_security=True):
        if check_security:
            request = check_request()
            permission_checker = request.has_permission
        else:
            permission_checker = granted_permission
        return [transition.transition_id
                for transition in sorted(self._get_transitions(MANUAL),
                                         key=lambda x: x.user_data.get('order', 999))
                if transition.condition(self, self.context) and
                    permission_checker(transition.permission, self.context)]

    def get_system_transition_ids(self):
        # ignore permission checks
        return [transition.transition_id
                for transition in sorted(self._get_transitions(SYSTEM),
                                         key=lambda x: x.user_data.get('order', 999))
                if transition.condition(self, self.context)]

    def get_fireable_transition_ids(self, check_security=True):
        return (self.get_manual_transition_ids(check_security) +
                self.get_system_transition_ids())

    def get_fireable_transition_ids_toward(self, state, check_security=True):
        result = []
        for transition_id in self.get_fireable_transition_ids(check_security):
            transition = self.wf.get_transition_by_id(transition_id)
            if transition.destination == state:
                result.append(transition_id)
        return result

    def get_automatic_transition_ids(self):
        return [transition.transition_id for transition in
                self._get_transitions(AUTOMATIC)]

    def has_automatic_transitions(self):
        return bool(self.get_automatic_transition_ids())

    def _get_transitions(self, trigger):
        # retrieve all possible transitions from workflow utility
        state = IWorkflowState(self.context)
        transitions = self.wf.get_transitions(state.state)
        # now filter these transitions to retrieve all possible
        # transitions in this context, and return their ids
        return [transition for transition in transitions if transition.trigger == trigger]


class WorkflowsVocabulary(UtilityVocabulary):
    """Workflows vocabulary"""

    interface = IWorkflow
    nameOnly = True

getVocabularyRegistry().register('PyAMS workflows', WorkflowsVocabulary)
